from typing import Any, List

from pydantic import BaseModel


class PageInput(BaseModel):
    sort_by: str = ""
    descending: bool = False
    page_size: int = 30
    page_start: int = 1


class PageOutput(BaseModel):
    rows_number: int = 0
    sort_by: str = ""
    descending: bool = False
    page_size: int = 30
    page_start: int = 1
    page_number: int = 1  # 页数
    rows: List[Any] = []



import datetime

from pydantic.main import ModelMetaclass

from db_hammer.page import PageInput
from db_hammer.util.date import date_to_str


def get_entity_fields(entity):
    """获取对象的所的字段"""
    dd = entity.__fields__
    _dict = {}
    ignore_columns = getattr(entity, "__ignore_columns__", None)
    for d in dd:
        if ignore_columns is not None and d in ignore_columns:
            continue
        if hasattr(entity, d):
            v = getattr(entity, d)
            if not d.startswith("_") \
                    and isinstance(v, (str, int, float, datetime.datetime, datetime.date)):
                _dict[d] = v
        else:
            _dict[d] = None
    return _dict


def tag_value(value):
    if value is None:
        return "%s", "null"
    else:
        if isinstance(value, str):
            return "%s", value
        elif isinstance(value, (datetime.datetime, datetime.date)):
            return "%s", date_to_str(value)
        elif isinstance(value, int):
            return "%d", value
        elif isinstance(value, float):
            return "%f", value
        else:
            raise Exception(f"不支持的类型:{value}-{type(value)}")


def insert_sql(entity) -> (str, []):
    dd = get_entity_fields(entity)
    values_tag = []
    values = {}
    table_name = getattr(entity, "__table_name__")
    for d in dd:
        value = getattr(entity, d)
        values_tag.append(f":{d}")  # 占位符
        values[d] = value
    return f"""INSERT INTO {table_name}({",".join(dd)}) VALUES ({",".join(values_tag)})""", values


def update_sql(entity, pass_null=False):
    primary_key = getattr(entity, "__primary_key__")
    table_name = getattr(entity, "__table_name__")
    where = ""
    values_tag = []
    values = {}
    dd = get_entity_fields(entity)
    for d in dd:
        value = getattr(entity, d)
        if value is None or value == "":
            if pass_null:
                continue
        values_tag.append(f"{d}=:{d}")
        values[d] = value
    if isinstance(primary_key, list):
        for k in primary_key:
            value = getattr(entity, k)
            values_tag.append(f"{k}=:{k}")
            values[k] = value
        if where == "":
            where += ' '.join(values_tag)
        else:
            where += ' AND '.join(values_tag)
    else:
        value = getattr(entity, primary_key)
        if where == "":
            where += f' {primary_key}=:{primary_key}'
        else:
            where += f' AND {primary_key}=:{primary_key}'

        values[primary_key] = value

    return f"""UPDATE {table_name} SET {",".join(values_tag)} WHERE{where}""", values


def delete_sql(entity):
    primary_key = getattr(entity, "__primary_key__")
    table_name = getattr(entity, "__table_name__")
    where = ""
    values_tag = []
    values = {}
    if isinstance(primary_key, list):
        for k in primary_key:
            value = getattr(entity, k)
            values_tag.append(f"{k}=:{k}")
            values[k] = value
            where += ' AND '.join(values_tag)
    else:
        value = getattr(entity, primary_key)
        where += f' {primary_key}=:{primary_key}'
        values[primary_key] = value
    return f"""DELETE FROM {table_name} WHERE {where}""", values


def entity_list(dict_list, cls) -> list:
    cols = get_entity_fields(cls)
    result = []
    for record in dict_list:
        en = {}
        for c in cols:
            en[c] = record[str(c).upper()]
        if type(cls.__class__) != ModelMetaclass:
            entity = cls(**en)
        else:
            entity = cls.__class__(**en)
        result.append(entity)
    return result


def where_entity(entity, rel="AND"):
    dd = get_entity_fields(entity)
    where = ""
    values = {}
    for d in dd:
        value = getattr(entity, d)
        if value is not None:
            # 字符串为空时，不加条件
            if isinstance(value, str) and value == "":
                continue
            if where == "":
                where += f' WHERE {d}=:{d}'
            else:
                where += f' {rel} {d}=:{d}'
            values[d] = value
    return where, values


def where_like_entity(entity, rel="AND"):
    dd = get_entity_fields(entity)
    where = ""
    values = {}
    for d in dd:
        value = getattr(entity, d)
        if value is not None:
            # 字符串为空时，不加条件
            if isinstance(value, str) and value == "":
                continue
            if where == "":
                where += f' WHERE {d} LIKE :{d}'
            else:
                where += f' {rel} {d} LIKE :{d}'
            values[d] = f"%{value}%"
    return where, values


def order_by_pagination(entity, pagination: PageInput):
    dd = get_entity_fields(entity)
    sortBy = pagination.sort_by
    descending = pagination.descending
    if sortBy != "" and sortBy in dd.keys():
        return f" ORDER BY {sortBy} {'' if descending else 'DESC'}"
    return ""


def get_entity_primary_key(entity):
    primary_key = getattr(entity, "__primary_key__")
    if isinstance(primary_key, list):
        pk_value = {}
        for pk in primary_key:
            pk_value[pk] = getattr(entity, pk)
        return primary_key, pk_value
    else:
        pk_value = getattr(entity, primary_key)
        return primary_key, pk_value



import datetime
import decimal
import gzip
import logging
import os

from db_hammer.util.date import date_to_str


def get_headers(cursor):
    col_names = cursor.description
    cc = []
    for c in col_names:
        cc.append(c[0])
    return cc


def count_rows(cursor, sql, log=None):
    log = log or logging.getLogger(__name__)
    count_sql = f"SELECT COUNT(0) FROM ({sql}) tmp_count"
    log.debug("execute sql:" + count_sql.replace("\n", " "))
    cursor.execute(count_sql)
    data = cursor.fetchone()
    log.debug("fetch rows:" + str(len(data)))
    num = data[0]
    return int(num)


def start(cursor, sql, bachSize, PACK_SIZE, path, file_mode, add_header, CSV_SPLIT, CSV_FIELD_CLOSE, encoding,
          callback, log=None):
    log = log or logging.getLogger(__name__)
    try:
        os.makedirs(path, exist_ok=True)
        log.info(f"export path：%s" % path)
        cursor.execute(sql)
        col_names = get_headers(cursor)
        csv_data = cursor.fetchmany(int(bachSize))
        file_i = 1
        CACHE_COUNT = PACK_SIZE
        CACHE_ROWS = 0
        write_f = None
        c_count = 0
        total = 0
        if callback is not None:
            total = count_rows(cursor, sql, log)
            callback(c_count, total)
        while len(csv_data) > 0:
            log.debug("export row::" + str(c_count))
            CACHE_ROWS += len(csv_data)
            if CACHE_ROWS < CACHE_COUNT and c_count > 0:
                file_i, file_name, write_f = write_file(col_names, csv_data, path, file_i, False, write_f, file_mode,
                                                        add_header, CSV_SPLIT, CSV_FIELD_CLOSE, encoding)
            else:
                file_i, file_name, write_f = write_file(col_names, csv_data, path, file_i, True, write_f, file_mode,
                                                        add_header, CSV_SPLIT, CSV_FIELD_CLOSE, encoding)
                CACHE_ROWS = 0
            c_count += len(csv_data)
            csv_data = cursor.fetchmany(int(bachSize))
            if callback is not None:
                callback(c_count, total)
        if write_f is not None:
            write_f.flush()
            write_f.close()
        log.info(f"export end, total row::%d" % c_count)
    except Exception as e:
        log.exception(e)
    finally:
        pass


def write_file(col_names, csv_data, path, file_i, new, f, file_mode, add_header, CSV_SPLIT, CSV_FIELD_CLOSE,
               encoding):
    add = False
    if file_mode == 'gz':
        file_name = f"{str(file_i).rjust(6, '0')}.gz"
    elif file_mode == 'csv':
        file_name = f"{str(file_i).rjust(6, '0')}.csv"
    else:
        file_name = f"{str(file_i).rjust(6, '0')}.txt"
    if len(csv_data) == 0:
        return file_i, None, None
    if new:
        if f is not None:
            f.flush()
            f.close()
        file_i += 1
        add = True
        if file_mode == 'gz':
            f = gzip.open(f"{path}/{file_name}", 'wb+')
        else:
            f = open(f"{path}/{file_name}", 'wb+')
    __run_rows(rows=csv_data,
               header=col_names,
               file=f,
               add_header=add and add_header,
               CSV_SPLIT=CSV_SPLIT,
               CSV_FIELD_CLOSE=CSV_FIELD_CLOSE,
               encoding=encoding
               )
    return file_i, file_name, f


def __run_rows(rows, header, add_header=False, file=None, CSV_FIELD_CLOSE='"', CSV_SPLIT=",", encoding="utf-8"):
    for r in rows:
        fields = r
        field_map = {}
        for i in range(0, len(header)):
            c = header[i]
            v = fields[i]
            if isinstance(v, datetime.datetime) or isinstance(v, datetime.date):
                if v is None:
                    field_map[c] = ""
                else:
                    field_map[c] = date_to_str(v)
            elif isinstance(v, int):
                if v is None:
                    field_map[c] = '0'
                else:
                    field_map[c] = str(v)
            elif isinstance(v, decimal.Decimal):
                if v is None:
                    field_map[c] = '0'
                else:
                    field_map[c] = str(round(v, 4))
            elif isinstance(v, float):
                if v is None:
                    v = '0'
                field_map[c] = str(v)
            elif isinstance(v, str):
                # 空处理
                if v is None:
                    field_map[c] = ""
                else:
                    # 回车换行处理
                    field_map[c] = str(v).replace("\n", "").replace(CSV_FIELD_CLOSE, ""). \
                        replace(CSV_SPLIT, "")
            else:
                if v is None:
                    field_map[c] = ""
                else:
                    raise Exception(f"[{c}]--NO SUPPORT TYPE--[{type(v)}]")
        if add_header:
            row_str = CSV_FIELD_CLOSE + f'{CSV_FIELD_CLOSE}{CSV_SPLIT}{CSV_FIELD_CLOSE}'.join(header) + CSV_FIELD_CLOSE
            line = (row_str + '\n').encode(encoding)
            add_header = False
            file.write(line)
        row_str = CSV_FIELD_CLOSE + f'{CSV_FIELD_CLOSE}{CSV_SPLIT}{CSV_FIELD_CLOSE}'.join(
            field_map.values()) + CSV_FIELD_CLOSE
        line = (row_str + '\n').encode(encoding)
        file.write(line)



# import platform
#
# import jpype
#
#
# class HurbaoTool:
#     jvmPath = jpype.getDefaultJVMPath()
#     jarPath = "-Djava.class.path=hrb-common_jm.jar"
#     if platform.system().lower() == 'windows':
#         jvmPath = r'C:\Program Files\Java\jdk1.8.0_131\jre\bin\server\jvm.dll'
#         jarPath = "-Djava.class.path=D:\\hrb-common_jm.jar"
#     jpype.startJVM(jvmPath, jarPath)
#     jpype.java.lang.System.out.println("hello world!")
#     DefaultSeedAES = jpype.JClass("com.hurbao.utils.secret.DefaultSeedAES")



import socket


def is_inuse(ip, port):
    """端口是否被占用"""
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(3)
        s.connect((ip, int(port)))
        s.shutdown(2)
        return True
    except:
        return False


def get_random_port(ip):
    """根据IP获取一个随机端口（15000~20000）"""
    import random
    times = 0
    max_times = 50
    port = random.randint(15000, 20000)
    while is_inuse(ip, port) and times < max_times:
        port = random.randint(15000, 20000)
        times += 1
    if times > max_times:
        Exception("端口号获取失败")
    return port


def get_pc_name_ip(host):
    """获取当前IP与主机名 返回:(ip,name)"""
    name = socket.getfqdn(socket.gethostname())
    s = None
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect((host.split(":")[0], int(host.split(":")[1])))
        ip = s.getsockname()[0]
    finally:
        s.close()
    return name, ip


def recv_end(the_socket, SOCKET_END_TAG):
    """通过寻找接收的协议数据中的尾标识字符串，获取完整的数据的数据报文"""
    total_data = []
    while True:
        data = the_socket.recv(8192)
        if SOCKET_END_TAG in data:
            total_data.append(data[:data.find(SOCKET_END_TAG)])
            break
        total_data.append(data)
        if len(total_data) > 1:
            # check if end_of_data was split
            last_pair = total_data[-2] + total_data[-1]
            if SOCKET_END_TAG in last_pair:
                total_data[-2] = last_pair[:last_pair.find(SOCKET_END_TAG)]
                total_data.pop()
                break
    if len(total_data) == 0:
        return None
    return b''.join(total_data)



import uuid

from db_hammer.base import BaseConnection
from db_hammer.csv import start as csv_start

try:
    import psycopg2
except ImportError:
    print("cant import psycopg2")
    print("===> pip3 install psycopg2")
    raise Exception("import psycopg2 error")


class PostgreSQLConnection(BaseConnection):

    def __init__(self, **kwargs):
        if kwargs.get("host", None) is None:
            raise Exception("host")
        if kwargs.get("user", None) is None:
            raise Exception("user")
        if kwargs.get("db_name", None) is None:
            raise Exception("db_name")
        if kwargs.get("pwd", None) is None:
            raise Exception("pwd")
        port = kwargs.get("port", 5432)

        super().__init__(**kwargs)
        self.conn = psycopg2.connect(database=kwargs["db_name"],
                                     user=kwargs["user"],
                                     password=kwargs["pwd"],
                                     host=kwargs["host"],
                                     port=port)

        self.cursor = self.conn.cursor()

    def db_data_type_mapping(self):
        return {
            "VARCHAR,CHAR,TEXT": str,
            "DECIMAL,FLOAT4,FLOAT8,MONEY": float,
            "INT2,INT4,INT8,SERIAL2,SERIAL4,SERIAL8,BOOLEAN": int,
            "DATE,TIMESTAMP,TIME": "to_date('${VALUE}','yyyy-mm-dd,hh24:mi:ss')",
        }

    def db_column_sql(self, table_name):
        '''
            COLUMN_NAME : 名称
            DATA_TYPE ： 类型
            COMMENTS ： 备注
            DATA_LENGTH ： 长度
            DATA_PRECISION ：精度
        :param table_name:
        :return:
        '''
        return f"""
          SELECT a.attnum,
                   a.attname AS COLUMN_NAME,
                   t.typname AS DATA_TYPE,
                   a.attlen AS DATA_LENGTH,
                   a.atttypmod AS lengthvar,
                   a.attnotnull AS notnull,
                   b.description AS COMMENTS
            FROM pg_class c,
                 pg_attribute a
                     LEFT OUTER JOIN pg_description b ON a.attrelid=b.objoid AND a.attnum = b.objsubid,
                 pg_type t
            WHERE c.relname = '{table_name}'
              and a.attnum > 0
              and a.attrelid = c.oid
              and a.atttypid = t.oid
            ORDER BY a.attnum
            """

    def convert_str(self, s: str):
        return s.replace("'", "''")

    def export_data_file(self, sql, dir_path, file_mode="gz", pack_size=500000, bachSize=10000, add_header=True,
                         data_split_chars=',',
                         data_close_chars='"', encoding="utf-8", outingCallback=None):
        """导出数据文件
        @:param sql 导出时的查询SQL
        @:param dir_path 导出的数据文件存放目录
        @:param file_mode 导出文件格式：txt|gz|csv
        @:param add_header 数据文件是否增加表头
        @:param pack_size  每个数据文件大小，默认为50万行，强烈建议分割数据文件，单文件写入速度会越来越慢
        @:param bachSize   游标大小
        @:param data_split_chars 每条数据字段分隔字符,csv文件默认为英文逗号
        @:param data_close_chars 每条数据字段关闭字符,csv文件默认为英文双引号
        @:param encoding 文件编码格式，默认为utf-8
        @:param outingCallback 导出过程中的回调方法
        """
        # 要使用服务器游标，本地游标会内存溢出
        cursor = self.conn.cursor(name="export_data_" + str(uuid.uuid1()))
        csv_start(cursor=cursor,
                  sql=sql,
                  path=dir_path,
                  bachSize=bachSize,
                  PACK_SIZE=pack_size,
                  file_mode=file_mode,
                  add_header=add_header,
                  CSV_SPLIT=data_split_chars,
                  CSV_FIELD_CLOSE=data_close_chars,
                  encoding=encoding,
                  callback=outingCallback,
                  log=self.log)
        cursor.close()



import sqlite3
from db_hammer.base import BaseConnection


class Sqlite3Connection(BaseConnection):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        database = kwargs.get("database", None)
        if database is None:
            raise Exception("database")
        del kwargs["database"]
        self.conn = sqlite3.connect(database=database, **kwargs)
        self.cursor = self.conn.cursor()




from db_hammer.base import BaseConnection

try:
    import Cython
except ImportError:
    print("cant import Cython")
    print("===> pip3 install Cython")
    raise Exception("import Cython error")
try:
    import pymssql
except ImportError:
    print("cant import pymssql")
    print("===> pip3 install pymssql")
    raise Exception("import pymssql error")


# pip install Cython
# pip install pymssql
class MsSQLConnection(BaseConnection):
    def __init__(self, **kwargs):
        if kwargs.get("host", None) is None:
            raise Exception("host")
        if kwargs.get("user", None) is None:
            raise Exception("user")
        if kwargs.get("db_name", None) is None:
            raise Exception("db_name")
        if kwargs.get("pwd", None) is None:
            raise Exception("pwd")
        port = kwargs.get("port", 1433)
        charset = kwargs.get("charset", "utf8")
        tds_version = kwargs.get("tds_version", "7.0")
        super().__init__(**kwargs)
        self.conn = pymssql.connect(server=kwargs["host"], user=kwargs["user"],
                                    password=kwargs["pwd"],
                                    port=port,
                                    charset=charset,
                                    tds_version=tds_version,
                                    database=kwargs["db_name"])
        self.cursor = self.conn.cursor()



import os
import re


def get_dir_files(path: str, mode='F', absolute=False, filter_re=""):
    """获取目录下所有文件
    :param path: 获取路径
    :arg mode
        F   当前目录下所有文件
        D   当前目录下所有文件夹
        FD  当前目录下所有文件和目录
        AD  当前目录和子目录
        AF  当前文件和子目录文件
        AFD 当前目录的子目录所有目录和文件
    :arg absolute 是否全路径
    :arg filter_re 正则正则
    """
    if not os.path.exists(path):
        raise Exception("文件目录不存在:" + path)

    if not os.path.isdir(path):
        raise Exception("不是目录：" + path)
    aa = []
    if mode == "F" or mode == "D" or mode == "FD":
        for root, dirs, files in os.walk(top=path, topdown=False):
            if mode == "F":
                if absolute:
                    ff = []
                    for f in files:
                        if root == path and re.match(filter_re, os.path.join(root, f)) is not None:
                            ff.append(os.path.join(root, f))
                    files = ff
                aa += files
            if mode == "D":
                if os.path.isdir(root) and re.match(filter_re, root) is not None:
                    aa.append(root)
            if mode == "FD" and re.match(filter_re, root) is not None:
                aa.append(root)
        return aa
    if mode == "AF" or mode == "AD" or mode == "AFD":
        for root, dirs, files in os.walk(top=path, topdown=True):
            if mode == "AF":
                if absolute:
                    ff = []
                    for f in files:
                        if re.match(filter_re, os.path.join(root, f)) is not None:
                            ff.append(os.path.join(root, f))
                    files = ff
                aa += files
            if mode == "AD":
                if os.path.isdir(root):
                    if re.match(filter_re, root) is not None:
                        aa.append(root)
            if mode == "AFD":
                if re.match(filter_re, root) is not None:
                    aa.append(root)
        return aa


if __name__ == '__main__':
    print(get_dir_files(path="/Users/lawman/", absolute=True, mode="F", filter_re=".*msi$"))
    # print(re.match("", "1234456"))



import os
import sys

from db_hammer.base import BaseConnection

try:
    import cx_Oracle
except ImportError:
    print("cant import cx_Oracle")
    print("===> pip3 install cx_Oracle")
    raise Exception("import cx_Oracle error")


class OracleConnection(BaseConnection):

    def __init__(self, **kwargs):
        if kwargs.get("host", None) is None:
            raise Exception("host")
        if kwargs.get("user", None) is None:
            raise Exception("user")
        if kwargs.get("db_name", None) is None:
            raise Exception("db_name")
        if kwargs.get("pwd", None) is None:
            raise Exception("pwd")
        port = kwargs.get("port", 1521)
        nlsLang = kwargs.get("nlsLang", "SIMPLIFIED CHINESE_CHINA.AL32UTF8")
        super().__init__(**kwargs)
        CONNECT_TNS = False
        # NLS_LANG = 'SIMPLIFIED CHINESE_CHINA.ZHS16GBK'
        # NLS_LANG = 'SIMPLIFIED CHINESE_CHINA.AL32UTF8'
        os.environ['NLS_LANG'] = nlsLang
        if kwargs.get("oracleHome", None) is not None:
            oracleHome = str(kwargs["oracleHome"]).lower().replace("oci.dll", "").replace("OCI.Dll", "")
            sys.path.append(oracleHome)
            os.environ['PATH'] = oracleHome
            os.environ['ORACLE_HOME'] = oracleHome
        self.conn = None
        if not CONNECT_TNS:
            try:
                self.conn = cx_Oracle.connect(kwargs["user"], kwargs["pwd"],
                                              f'{kwargs["host"]}:{port}/{kwargs["db_name"]}',
                                              encoding="UTF-8", nencoding="UTF-8")
            except Exception as e:
                if "12514" in str(e):
                    CONNECT_TNS = True
                else:
                    raise e
        if CONNECT_TNS:
            try:
                print("尝试TNS..")
                dsn = cx_Oracle.makedsn(kwargs["host"], port, service_name=kwargs["db_name"])
                self.conn = cx_Oracle.connect(kwargs["user"], kwargs["pwd"], dsn=dsn, encoding="UTF-8",
                                              nencoding="UTF-8")
            except Exception as e:
                raise e
        self.cursor = self.conn.cursor()

    def db_data_type_mapping(self):
        return {
            "VARCHAR2,NVARCHAR2,CHAR": str,
            "FLOAT,NUMBER": float,
            "INTEGER": int,
            "DATE,TIMESTAMP": "to_date('${VALUE}','yyyy-mm-dd,hh24:mi:ss')",
        }

    def db_column_sql(self, table_name):
        '''
            COLUMN_NAME : 名称
            DATA_TYPE ： 类型
            COMMENTS ： 备注
            DATA_LENGTH ： 长度
            DATA_PRECISION ：精度
        :param table_name:
        :return:
        '''
        return f"""
            SELECT TC.COLUMN_ID,
               TC.NULLABLE,
               TC.COLUMN_NAME,
               DATA_TYPE,
               DATA_LENGTH,
               DECODE(DATA_TYPE, 'NUMBER', DATA_PRECISION || ',' || DATA_SCALE, DATA_LENGTH) DATA_PRECISION,
               CC.COMMENTS,
               TC.DATA_DEFAULT,
                CS.COMMENTS TABLE_COMMENTS,
               pk.COLUMN_NAME PK_TAG
        FROM USER_TAB_COLUMNS TC
                 LEFT JOIN USER_COL_COMMENTS CC
                           ON TC.TABLE_NAME = CC.TABLE_NAME AND TC.COLUMN_NAME = CC.COLUMN_NAME
                 left join (
            select col.COLUMN_NAME
            from user_constraints con,
                 user_cons_columns col
            where con.constraint_name = col.constraint_name
              and con.constraint_type = 'P'
              and col.table_name = '{table_name}'
        ) PK on tc.COLUMN_NAME = pk.COLUMN_NAME
        left join user_tab_comments CS
                   ON TC.TABLE_NAME = CS.TABLE_NAME
        WHERE TC.TABLE_NAME = '{table_name}'
        """

    def convert_str(self, s: str):
        return s.replace("'", "''")



import poplib
import smtplib
from email.header import decode_header
from email.mime.text import MIMEText
from email.parser import Parser
from email.utils import parseaddr


class SimEmail:
    def __init__(self, username, password, sender,mtpserver='smtp.qq.com', popServer="pop.qq.com"):
        self.username = username
        self.password = password
        self.sender = sender
        self.mtpserver = mtpserver
        self.popserver = popServer

    def sendMail(self, subject, text, receiver):
        smtpserver = self.mtpserver
        sender = self.sender
        username = self.username
        password = self.password
        msg = MIMEText(text, 'html', 'utf-8')
        msg['From'] = sender
        msg['To'] = receiver
        msg['Subject'] = subject

        smtp = smtplib.SMTP()
        smtp.connect(smtpserver)
        smtp.login(username, password)
        smtp.sendmail(sender, receiver, msg.as_string())
        smtp.quit()

    def getMail(self, index=None, isDelete=False):
        # 连接到POP3服务器:
        server = poplib.POP3_SSL(self.popserver)

        # 可以打开或关闭调试信息:
        # server.set_debuglevel(1)

        # 可选:打印POP3服务器的欢迎文字:
        # print(server.getwelcome())

        # 身份认证:
        server.user(self.username)
        server.pass_(self.password)

        # stat()返回邮件数量和占用空间:
        # print('Messages: %s. Size: %s' % server.stat())

        # list()返回所有邮件的编号:
        resp, mails, octets = server.list()

        # 可以查看返回的列表类似['1 82923', '2 2184', ...]
        # print(mails)

        # 获取最新一封邮件, 注意索引号从1开始:
        if index is None:
            index = len(mails)
        resp, lines, octets = server.retr(index)

        # lines存储了邮件的原始文本的每一行,
        # 可以获得整个邮件的原始文本:
        msg_content = bytes('\r\n', encoding='utf-8').join(lines)

        # 稍后解析出邮件:
        msg = Parser().parsestr(str(msg_content, encoding='utf-8'))

        result = self.print_info(msg)

        result["mail_list"] = mails

        # 可以根据邮件索引号直接从服务器删除邮件:
        if isDelete:
            server.dele(index)

        # 关闭连接:
        server.quit()
        return result

    # 猜测编码格式
    def guess_charset(self, msg):
        charset = msg.get_charset()
        if charset is None:
            content_type = msg.get('Content-Type', '').lower()
            pos = content_type.find('charset=')
            if pos >= 0:
                charset = content_type[pos + 8:].strip()
        return charset

    # 解码
    def decode_str(self, s):
        value, charset = decode_header(s)[0]
        if charset:
            value = value.decode(charset)
        return value

    # 解析邮件内容
    def print_info(self, msg, indent=0):
        result = {}
        if indent == 0:
            for header in ['From', 'To', 'Subject']:
                value = msg.get(header, '')
                if value:
                    if header == 'Subject':
                        value = self.decode_str(value)
                    else:
                        hdr, addr = parseaddr(value)
                        name = self.decode_str(hdr)
                        value = u'%s <%s>' % (name, addr)
                # print('%s%s: %s' % ('  ' * indent, header, value))
                result[header] = value
        if (msg.is_multipart()):
            parts = msg.get_payload()
            for n, part in enumerate(parts):
                # print('%spart %s' % ('  ' * indent, n))
                # print('%s--------------------' % ('  ' * indent))
                self.print_info(part, indent + 1)
        else:
            content_type = msg.get_content_type()
            if content_type == 'text/plain' or content_type == 'text/html':
                content = msg.get_payload(decode=True)
                charset = self.guess_charset(msg)
                if charset:
                    content = content.decode(charset)
                # print('%sText: %s' % ('  ' * indent, content + '...'))
                result["Content"] = content
            else:
                # print('%sAttachment: %s' % ('  ' * indent, content_type))
                result["Attachment"] = content_type
        return result


if __name__ == '__main__':
    pass


def table(name=None, primary_key="id", column_map=None):
    """数据据保存的表名"""
    def decorate(clazz):
        setattr(clazz, "__table_name__", clazz.__name__ if name is None else name)
        setattr(clazz, "__primary_key__", primary_key)
        setattr(clazz, "__column_map__", None if column_map is None else column_map)
        return clazz
    return decorate



import datetime
import logging
import re
from enum import Enum

from db_hammer.csv import start as csv_start
from db_hammer.entity_util import update_sql, insert_sql, delete_sql, entity_list, where_entity, where_like_entity, \
    order_by_pagination, get_entity_primary_key
from db_hammer.page import PageOutput, PageInput
from db_hammer.sql_exception import SqlException, FetchRowsException, ExistException
from db_hammer.util.date import date_to_str


class DataType(Enum):
    DateNull = -2177774
    DataPass = -2177771


class BaseConnection(object):
    def __init__(self, **kwargs):
        """
        :param kwargs:
        """
        self.log = kwargs.get("log", logging.getLogger(__name__))
        self.Db_NAME = kwargs.get("db_name")
        self.conn = None  # type:Connect
        self.cursor = None  # type:cursor
        self.table_column_cache = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def select_page_size(self, sql: str, params: dict = None, page_size=50) -> (int, int):
        """
        根据sql获取 (页数,行数)
        :param params:
        :param sql:
        :param page_size: 分页大小
        :return: 页数
        """

        count_sql = "select COUNT(0) from ( %s ) temp_count" % sql
        count_sql = self.sql_params(count_sql, params)
        self.log.debug("执行SQL:" + count_sql.replace("\n", " "))
        self.log.debug("参数:" + str(params))
        self.cursor.execute(count_sql, params)
        data = self.cursor.fetchone()
        count_rows = data[0]
        self.log.debug("影响行数:" + str(len(data)))
        num = count_rows // page_size
        if count_rows > page_size * num:
            num = num + 1
        return num, count_rows

    def sql_params(self, sql: str, params: dict):
        if params is None:
            return sql
        ps = re.findall('(?<!\w):\w+', sql)
        start = 0
        for p in ps:
            if p[1:] not in params or len(p) == 1:
                raise SqlException(f"SQL语句中参数[{p}]没有传值")
            start = sql.find(p, start)
            param = "%(" + p[1:] + ")s"
            sql = sql[:start] + param + sql[start + len(p):]
            start += len(param)
        return sql

    def select_page_list(self, sql: str, page_size=50, page_start=1, **kwargs) -> list:
        """
        获取分页列表数据,列表方式返回
        :param sql:
        :param page_size: 分页大小
        :param page_start: 开始页 （从1开始）

        **kwargs
        page_end: 结束页 defalut：page_start+1
        add_headers: 是否添加表头 defalut:False
        :return:
        """
        params = kwargs.get("params", None)
        page_end = kwargs["page_end"]
        if page_end is None:
            page_end = page_start + 1
        add_headers = kwargs["add_headers"]
        if add_headers is None:
            add_headers = False

        start = page_size * (page_start - 1)
        end = page_size * (page_end - page_start)
        page_sql = """SELECT * FROM ( %s ) temp_page LIMIT %d,%d """ % (sql, start, end)
        self.log.debug("执行SQL:" + page_sql.replace("\n", " "))
        page_sql = self.sql_params(page_sql, params)
        self.log.debug("参数:" + str(params))
        self.cursor.execute(page_sql, params)
        data = self.cursor.fetchall()
        self.log.debug("影响行数:" + str(len(data)))
        if add_headers:
            col_names = self.cursor.description
            list(data).insert(col_names, 0)
        return data

    def select_dict_page_list(self, sql: str, page_size=50, page_start=1, **kwargs) -> list:
        """
        获取分页列表数据，,字典形式返回
        :param sql:
        :param page_size:
        :param page_start:
        :param page_end:
        :return:
        """

        page_end = kwargs.get("page_end", page_start + 1)
        add_headers = kwargs.get("add_headers", False)
        params = kwargs.get("params", None)
        data = self.select_page_list(sql, page_size, page_start, params=params,
                                     page_end=page_end,
                                     add_headers=add_headers)
        col_names = self.cursor.description
        return self._data_to_map(col_names, data)

    def select_value(self, sql: str, params=None) -> str:
        """
        获取第一列第一行的值
        :param sql:
        :return:
        """
        self.log.debug("执行SQL:" + sql.replace("\n", " "))
        _sql = self.sql_params(sql, params)
        self.cursor.execute(_sql, params)
        data = self.cursor.fetchone()
        if data is None:
            return None
        return str(data[0])

    def select_list(self, sql: str, params=None) -> list:
        """
        获取列表数据,列表方式返回
        :param sql:
        :return:
        """
        self.log.debug("执行SQL:" + sql.replace("\n", " "))
        self.log.debug("参数:" + str(params))
        _sql = self.sql_params(sql, params)
        self.cursor.execute(_sql, params)
        data = self.cursor.fetchall()
        self.log.debug("影响行数:" + str(len(data)))
        return data

    def select_dict_list(self, sql: str, params=None) -> list:
        """
        获取查询语句列表数据，以字典方式返回: List<dict>
        :param params:
        :param sql:
        :return:
        """
        self.log.debug("执行SQL:" + sql.replace("\n", " "))
        self.log.debug("参数:" + str(params))
        _sql = self.sql_params(sql, params)
        self.cursor.execute(_sql, params)
        data = self.cursor.fetchall()
        self.log.debug("影响行数:" + str(len(data)))
        col_names = self.cursor.description
        return self._data_to_map(col_names, data)

    def select_dict(self, sql: str, params=None) -> dict:
        """
        获取字典数据，以字典方式返回: dict
        :param params:
        :param sql:
        :return:
        """
        data = self.select_dict_list(sql, params)
        if data is None or len(data) == 0:
            return None
        return data[0]

    @staticmethod
    def _data_to_map(col_names, data):
        r_list = []
        for k in range(len(data)):
            row = {}
            for i in range(len(col_names)):
                row[col_names[i][0]] = data[k][i]
            r_list.append(row)
        return r_list

    def execute(self, sql: str, params=None) -> int:
        """
        执行sql,并返回影响行数
        :param params:
        :param sql:
        :return:
        """
        _sql = self.sql_params(sql, params)
        self.log.debug("执行SQL:" + sql.replace("\n", " "))
        self.log.debug("参数:" + str(params))
        self.cursor.execute(_sql, params)
        i = self.cursor.rowcount
        self.log.debug("影响行数:" + str(i))
        return i

    def close(self):
        self.conn.close()
        # self.log.debug("关闭连接")

    def rollback(self):
        """回滚事务"""
        # self.log.debug("回滚事务开始")
        self.conn.rollback()
        self.log.debug("回滚事务完成")

    def commit(self):
        """提交事务，如果有对数据库进更新，需要手动提交事务
        """
        # self.log.debug("提交事务开始")
        self.conn.commit()
        self.log.debug("提交事务完成")
        return

    def db_data_type_mapping(self):
        """数据库与Python类型映射
            key: 数据库类型
            value: Py类型
            默认为MySQL数据库
        """

        return {
            "CHAR,VARCHAR,TINYBLOB,TINYTEXT,BLOB,TEXT,MEDIUMBLOB,MEDIUMTEXT,LONGBLOB,LONGTEXT": str,
            "TINYINT,SMALLINT,MEDIUMINT,INT,INTEGER,BIGINT": float,
            "FLOAT,DOUBLE,DECIMAL": int,
            "DATE,TIME,YEAR,DATETIME,TIMESTAMP": "STR_TO_DATE('${VALUE}', '%Y-%m-%d %H:%i:%s')",
        }

    def __get_py_type(self, db_type):
        dtm = self.db_data_type_mapping()
        for dt in dtm.keys():
            dd = dt.split(",")
            if str(db_type).upper() in dd:
                return dtm[dt]
        raise Exception("不支持的数据库类：" + str(db_type))

    def db_column_sql(self, table_name):
        """
        获取数据列的SQL
        占位参数：
            ${TABLE_NAME} : 表名
            ${DB_NAME} : 连接的数据库名

        :return:
            COLUMN_NAME : 名称
            DATA_TYPE ： 类型
            COMMENTS ： 备注
            DATA_LENGTH ： 长度
            DATA_PRECISION ：精度
        """
        return f"""
            SELECT 
             COLUMN_NAME AS COLUMN_NAME,
            DATA_TYPE AS DATA_TYPE,
            COLUMN_COMMENT AS COMMENTS,
            CHARACTER_MAXIMUM_LENGTH AS DATA_LENGTH,
            NUMERIC_PRECISION AS DATA_PRECISION
             FROM INFORMATION_SCHEMA.COLUMNS T
             WHERE 
             TABLE_NAME = '{table_name}' AND TABLE_SCHEMA = '{self.Db_NAME}'
            """

    def gen_insert_dict_sql(self, dict_data: dict, table_name: str):
        """
        通过字典保存数据，字典的key为表列表
        :param dict_data:
        :param table_name:
        :return:
        """
        columns = self.__select_db_columns(table_name=table_name)
        values = {}
        if dict_data is not None:
            for c in columns:
                column_name = c.get("COLUMN_NAME", None) if c.get("COLUMN_NAME", None) is not None else c["column_name"]
                data_type = c.get("DATA_TYPE", None) if c.get("DATA_TYPE", None) is not None else c["data_type"]
                values[column_name] = self.__get_column_value(data_type=data_type, value=dict_data[column_name])
        else:
            return None
        sql = f"""INSERT INTO {self.Db_NAME}.{table_name} ({','.join(values.keys())}) VALUES ({','.join(values.values())})"""
        return sql

    def __get_column(self, name, columns):
        for c in columns:
            column_name = c.get("COLUMN_NAME", None) if c.get("COLUMN_NAME", None) is not None else c["column_name"]
            if column_name == name:
                return c
        return None

    def __select_db_columns(self, table_name):
        cache = self.table_column_cache.get(table_name, None)
        if cache is not None:
            return cache
        else:
            columns = self.select_dict_list(sql=self.db_column_sql(table_name=table_name))
            self.table_column_cache[table_name] = columns
            return columns

    def gen_update_dict_sql(self, dict_data: dict, table_name: str, where: str, not_update_colunms=[]):
        """
        根据字典生成Update语句
        :param dict_data:  列新字典
        :param table_name: 表名
        :param where: 带上where条件
        :return:
        """
        columns = self.__select_db_columns(table_name=table_name)
        values = {}
        if dict_data is not None:
            for column_name in dict_data.keys():
                column = self.__get_column(column_name, columns)
                if column is not None:
                    data_type = column.get("DATA_TYPE", None) if column.get("DATA_TYPE", None) is not None else column[
                        "data_type"]
                    values[column_name] = self.__get_column_value(data_type=data_type, value=dict_data[column_name])
        else:
            return None
        sets = []
        for c in values.keys():
            if c not in not_update_colunms:
                sets.append(f"{c}={values[c]}")
        update_sql = f"UPDATE {self.Db_NAME}.{table_name} SET {','.join(sets)} " + where
        return update_sql

    def __get_column_value(self, data_type, value):
        """获取列的值"""
        if data_type is None:
            return f"'{self.convert_str(value)}'"
        t = self.__get_py_type(db_type=data_type)
        if value is None:
            return "null"
        elif t == str:
            return f"'{self.convert_str(value)}'"
        elif t in (int, float):
            return str(value)
        else:
            if isinstance(value, (datetime.datetime, datetime.date)):
                return t.replace("${VALUE}", date_to_str(value))
            else:
                return t.replace("${VALUE}", self.convert_str(value))

    def select_insert_sql(self, sql: str, table_name: str) -> list:
        """根据SQL查询数据并生成Insert语句"""
        records = self.select_dict_list(sql=sql)
        sql_list = []
        for r in records:
            sql = self.gen_insert_dict_sql(dict_data=r, table_name=table_name)
            sql_list.append(sql)
        return sql_list

    def convert_str(self, s: str):
        return str(s)

    def export_data_file(self, sql, dir_path, file_mode="txt", pack_size=500000, bachSize=10000, add_header=True,
                         data_split_chars=',',
                         data_close_chars='"', encoding="utf-8", outingCallback=None):
        """导出数据文件
        @:param sql 导出时的查询SQL
        @:param dir_path 导出的数据文件存放目录
        @:param file_mode 导出文件格式：txt|gz|csv
        @:param add_header 数据文件是否增加表头
        @:param pack_size  每个数据文件大小，默认为50万行，强烈建议分割数据文件，单文件写入速度会越来越慢
        @:param bachSize   游标大小
        @:param data_split_chars 每条数据字段分隔字符,csv文件默认为英文逗号
        @:param data_close_chars 每条数据字段关闭字符,csv文件默认为英文双引号
        @:param encoding 文件编码格式，默认为utf-8
        @:param outingCallback 导出过程中的回调方法
        """
        csv_start(cursor=self.cursor,
                  sql=sql,
                  path=dir_path,
                  bachSize=bachSize,
                  PACK_SIZE=pack_size,
                  file_mode=file_mode,
                  add_header=add_header,
                  CSV_SPLIT=data_split_chars,
                  CSV_FIELD_CLOSE=data_close_chars,
                  encoding=encoding,
                  callback=outingCallback,
                  log=self.log)

    def insert_entity(self, entity, commit=True):
        """传入实体保存到数据库"""
        sql, values = insert_sql(entity=entity)
        sql = self.sql_params(sql, values)
        i = self.execute(sql, values)
        primary_key = self.cursor.lastrowid
        if primary_key != 0:
            setattr(entity, get_entity_primary_key(entity)[0], primary_key)
        if i != 1:
            raise FetchRowsException(f"影响条数为{i}")
        if commit:
            self.commit()
        return entity

    def update_entity(self, entity, pass_null=False, commit=True, check_exist=False) -> None:
        if check_exist:
            es = self.select_entity_by_pk(entity=entity)
            if es is None:
                raise ExistException(f"找不到要更新的实体")
        sql, values = update_sql(entity=entity, pass_null=pass_null)
        self.log.debug(f"SQL:{sql}")
        self.log.debug(f"params:{values}")
        sql = self.sql_params(sql, values)
        i = self.execute(sql, values)
        if commit:
            self.commit()

    def delete_entity(self, entity, commit=True) -> int:
        sql, values = delete_sql(entity=entity)
        self.log.debug(sql)
        self.log.debug(f"params:{values}")
        sql = self.sql_params(sql, values)
        i = self.execute(sql, values)
        if commit:
            self.commit()
        return i

    def select_entity_list(self, cls=None, sql=None, params=None, like_entity=None, entity=None, rel="AND",
                           page: PageInput = None):
        if sql is None:
            if like_entity is not None:
                if cls is None:
                    cls = like_entity.__class__
                table_name = getattr(like_entity, "__table_name__")
            elif entity is not None:
                if cls is None:
                    cls = entity.__class__
                table_name = getattr(entity, "__table_name__")
            else:
                table_name = getattr(cls, "__table_name__")

            sql = "SELECT * FROM " + table_name
        if params is None:
            if like_entity is not None:
                where, values = where_like_entity(entity=like_entity, rel=rel)
                sql += where
                params = values
            elif entity is not None:
                where, values = where_entity(entity=entity, rel=rel)
                sql += where
                params = values
        if sql is None:
            raise Exception("查询SQL不能为空")
        if page is not None:
            pages, rowsNumber = self.select_page_size(sql=sql, params=params,
                                                      page_size=page.page_size)
            order_by = order_by_pagination(entity=cls, pagination=page)
            dict_list = self.select_dict_page_list(sql=f"{sql} {order_by}", params=params,
                                                   page_size=page.page_size,
                                                   page_start=page.page_start)
            values = entity_list(dict_list=dict_list, cls=cls)
            pageOut = PageOutput()
            pageOut.rows = values
            pageOut.page_start = page.page_start
            pageOut.page_size = page.page_size
            pageOut.page_number = pages
            pageOut.rows_number = rowsNumber
            pageOut.sort_by = page.sort_by
            pageOut.descending = page.descending
            return pageOut
        else:
            dict_list = self.select_dict_list(sql=sql, params=params)
            values = entity_list(dict_list=dict_list, cls=cls)
            return values

    def select_entity_first(self, cls=None, sql=None, params=None, like_entity=None, entity=None):
        ll = self.select_entity_list(cls, sql=sql, params=params, like_entity=like_entity, entity=entity)
        if ll is not None and len(ll) > 0:
            return ll[0]

    def select_entity_by_pk(self, entity, pk_value=None):
        primary_key, _pk_value = get_entity_primary_key(entity)
        table_name = getattr(entity, "__table_name__")
        _SQL = ""
        if isinstance(primary_key, list):
            for pk in primary_key:
                _SQL += f"""{pk}=:{pk} AND"""
            _SQL = _SQL[:-3]
        else:
            _SQL = f"""{primary_key}=:{primary_key}"""
        sql = f"""SELECT * FROM {table_name} WHERE {_SQL}"""
        if pk_value is None:
            pk_value = _pk_value

        params = {}
        if isinstance(pk_value, str):
            params[primary_key] = pk_value
        else:
            params = pk_value

        ll = self.select_entity_list(entity, sql=sql, params=params)
        if ll is not None and len(ll) > 0:
            return ll[0]
